#include "ipv4.h"

#include "../ipv4_config/ipv4_config.h"
#include "../ipv4_route_table/ipv4_route_table.h"
#include "../arp/arp.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <timerms.h>

/* Dirección IPv4 a cero: "0.0.0.0" */
ipv4_addr_t IPv4_ZERO_ADDR = { 0, 0, 0, 0 };

/* Multicast IP para recibir paquetes RIP cuando son enviados a multicast: "224.0.0.9" */
ipv4_addr_t RIP_IP_MULTICAST_ADDR = { 224, 0, 0, 9 };

#define IPv4_HEADER_LENGTH 20 /* Tamaño de la cabecera IP */
#define IP_VERSION_LON 0x45 /* Version y longitud IP: 4(versión) y 5(tamaño mínimo de cabecera, porque no utilizamos opciones)*/

/*
	Estructura de un paquete IPv4, según la RFC791: https://tools.ietf.org/html/rfc791

	0                   1                   2                   3
   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  |Version|  IHL  |Type of Service|          Total Length         |
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  |         Identification        |Flags|      Fragment Offset    |
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  |  Time to Live |    Protocol   |         Header Checksum       |
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  |                       Source Address                          |
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  |                    Destination Address                        |
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  |                    Options                    |    Padding    | (se obvia en nuestro caso)
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  |                        	Payload 						   |
  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

*/

struct ipv4_packet {
	uint8_t version_ihl; /* Versión + longitud de cabecera */
	uint8_t service; /* Indica la forma en la que se debe procesar el datagrama. */
	uint16_t total_length;  /* Logintud total del paquete IP */
	uint16_t id; /* Identificador */
	uint16_t flags_fragment; /* Flags + Fragment offset, en esta asignatura por defecto será 0 */
	uint8_t ttl; /* Tiempo de vida de la trama IP */
	uint8_t protocol; /* This field indicates the next level protocol used in the data portion of the internet datagram. */
	uint16_t checksum; /* La suma de comprobación es la suma de todas las palabras de 16 bits del encabezado (excluyendo sum) */
	ipv4_addr_t src_addr_ipv4; /* Dirección IP origen */
	ipv4_addr_t dst_addr_ipv4; /* Dirección IP destino */
	unsigned char payload [IPv4_MTU]; /* Datos de información, ETH_MTU - IP_HEADER_LENGTH */
};

struct ipv4_layer {
	eth_iface_t * iface;
	ipv4_addr_t addr;
	ipv4_addr_t netmask;
	ipv4_route_table_t * table;
};

/* void ipv4_addr_str ( ipv4_addr_t addr, char* str );
*
* DESCRIPCIÓN:
*   Esta función genera una cadena de texto que representa la dirección IPv4
*   indicada.
*
* PARÁMETROS:
*   'addr': La dirección IP que se quiere representar textualente.
*    'str': Memoria donde se desea almacenar la cadena de texto generada.
*           Deben reservarse al menos 'IPv4_STR_MAX_LENGTH' bytes.
*/

void ipv4_addr_str ( ipv4_addr_t addr, char* str )
{
  if (str != NULL)
  {
    sprintf(str, "%d.%d.%d.%d", addr[0], addr[1], addr[2], addr[3]);
  }
}

/* int ipv4_str_addr ( char* str, ipv4_addr_t addr );
*
* DESCRIPCIÓN:
*   Esta función analiza una cadena de texto en busca de una dirección IPv4.
*
* PARÁMETROS:
*    'str': La cadena de texto que se desea procesar.
*   'addr': Memoria donde se almacena la dirección IPv4 encontrada.
*
* VALOR DEVUELTO:
*   Se devuelve 0 si la cadena de texto representaba una dirección IPv4.
*
* ERRORES:
*   La función devuelve -1 si la cadena de texto no representaba una
*   dirección IPv4.
*/

int ipv4_str_addr ( char* str, ipv4_addr_t addr )
{
 int err = -1;

 if (str != NULL) {
   unsigned int addr_int[IPv4_ADDR_SIZE];
   int len = sscanf(str, "%d.%d.%d.%d",
                    &addr_int[0], &addr_int[1],
                    &addr_int[2], &addr_int[3]);

   if (len == IPv4_ADDR_SIZE) {
     int i;
     for (i=0; i<IPv4_ADDR_SIZE; i++) {
       addr[i] = (unsigned char) addr_int[i];
     }

     err = 0;
   }
 }

 return err;
}

/*
* uint16_t ipv4_checksum ( unsigned char * data, int len )
*
* DESCRIPCIÓN:
*   Esta función calcula el checksum IP de los datos especificados.
*
* PARÁMETROS:
*   'data': Puntero a los datos sobre los que se calcula el checksum.
*    'len': Longitud en bytes de los datos.
*
* VALOR DEVUELTO:
*   El valor del checksum calculado.
*/

uint16_t ipv4_checksum ( unsigned char * data, int len )
{
 int i;
 uint16_t word16;
 unsigned int sum = 0;

 /* Make 16 bit words out of every two adjacent 8 bit words in the packet
  * and add them up */
 for (i=0; i<len; i=i+2) {
   word16 = ((data[i] << 8) & 0xFF00) + (data[i+1] & 0x00FF);
   sum = sum + (unsigned int) word16;
 }

 /* Take only 16 bits out of the 32 bit sum and add up the carries */
 while (sum >> 16) {
   sum = (sum & 0xFFFF) + (sum >> 16);
 }

 /* One's complement the result */
 sum = ~sum;

 return (uint16_t) sum;
}

/*
 * ipv4_layer_t * ipv4_open(char* ipv4_config_file, char* ipv4_route_table_file);
 *
 * DESCRIPCIÓN:
 *   Esta función inicializa la tabla de rutas, abre el manejador de la
 *   interfaz Ethernet y obtiene nuestra la IPv4 y netmask de la máquina donde origen.
 *
 * PARÁMETROS:
 *    'ipv4_config': Nombre del fichero de configuración
 *    'ipv4_route_table': Nombre del fichero de tabla de rutas
 *
 * ERRORES:
 *   La función devuelve 'NULL' si se ha producido algún error.
 *
 * VALOR DEVUELTO:
 *   La función devuelve un puntero de tipo ipv4_layer_t con los datos proporcionados
 *	 por un fichero de configuración y una tabla de rutas.
 */

ipv4_layer_t * ipv4_open(char* ipv4_config_file, char* ipv4_route_table_file)
{
  ipv4_layer_t * ipv4_layer = malloc(sizeof(ipv4_layer_t));

  char ifname [IFACE_NAME_MAX_LENGTH];

	int read_config = ipv4_config_read(ipv4_config_file, ifname, ipv4_layer->addr, ipv4_layer->netmask);

	if (read_config == -1) {
    printf("Error al leer el fichero de configuración\n");
	  return NULL;
	}

	ipv4_layer->table = ipv4_route_table_create(); /* Reversamos memoria para la tabla de rutas, antes las leía pero no las guardaba */
	int read_tables = ipv4_route_table_read(ipv4_route_table_file, ipv4_layer->table);

	if (read_tables == -1) {
		printf("Error al leer el fichero de rutas\n");
		return NULL;
	}

	ipv4_layer->iface = eth_open(ifname);

	if(ipv4_layer->iface == NULL){
		printf("Error al abrir la interfaz Ethernet\n");
		return NULL;
	}

	return ipv4_layer;

}

/*
*	int ipv4_close(ipv4_layer_t * ipv4_layer )
*
* DESCRIPCIÓN:
*			Esta función cierra la interfaz utilizada por Ethernet,
*			libera la memoria asignada a la tabla de rutas y
*			libera la memoria reservada a la que apunta el puntero ipv4_layer.
*
* PARÁMETROS:
*    'ipv4_layer' : Puntero a un espacio de memoria de tamaño del tipo ipv4_layer_t
*									 donde se guardara temporalmente los datos para trabajar con dicha estructura
*									 Dicho espacio será liberado en esta función.
*
* VALOR DEVUELTO:
*   La función devuelve 0 si se ha cerrado correctamente, ó -1 en caso de error.
*
* ERRORES:
*   La función devuelve '-1' si se ha producido algún error.
*/

int ipv4_close ( ipv4_layer_t * ipv4_layer )
{
  ipv4_route_table_free (ipv4_layer->table);
  int err = eth_close (ipv4_layer->iface );
  free(ipv4_layer);

  if(err == -1)
  {
    printf("Error en ipv4_close.\n");
    return -1;
	}
	return 0;
}

/*
*	int ipv4_send
*	( ipv4_layer_t * ipv4, ipv4_addr_t ipv4_addr_dest,
*	uint8_t protocol, unsigned char *data, int data_length)
*
* DESCRIPCIÓN:
*   Esta función envía la información de la capa superior a la capa inferior (Ethernet)
*
* PARÁMETROS:
*		'ipv4_layer'			: Estructura ipv4_layer_t con los datos de la máquina que envía el paquete IP
*    'ipv4_addr_dest'	: IP destino
*    'protocol'				: Tipo de protocolo.
*    'data'						: Datos a enviar.
*    'length'					: Longitud de los datos que vamos a enviar.
*
* VALOR DEVUELTO:
*   El número de bytes de datos que han podido ser enviados.
*
* ERRORES:
*   La función devuelve '-1' si se ha producido algún error.
*/

int ipv4_send( ipv4_layer_t * ipv4_layer, ipv4_addr_t ipv4_addr_dest, uint8_t protocol, unsigned char *data, int data_length)
{
	int total_length = data_length + IPv4_HEADER_LENGTH;

	struct ipv4_packet packet;
	/* IP packet to send */
	packet.version_ihl = IP_VERSION_LON;
	packet.service = 0x00; /* Campo service de la cabecera IP */
	packet.total_length = htons(total_length);
	packet.id = htons(0x272F);  /*Poner un NIA*/
	packet.flags_fragment = 0x0000; /* Fragmentación de datagramas */
	packet.ttl = 64; /* Time to live para IPv4 */
	packet.protocol = protocol;
	packet.checksum = htons(0x0000);
	memcpy(packet.src_addr_ipv4, ipv4_layer->addr, IPv4_ADDR_SIZE);
  memcpy(packet.dst_addr_ipv4, ipv4_addr_dest, IPv4_ADDR_SIZE);
	memcpy(packet.payload, data, data_length);
	packet.checksum = htons(ipv4_checksum((unsigned char *)&packet, IPv4_HEADER_LENGTH));
	/* -IP packet */

	ipv4_route_t * best_route = ipv4_route_table_lookup (ipv4_layer->table, ipv4_addr_dest);
	mac_addr_t mac_dst = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	int arp = 0;

	if(best_route == NULL) {
	  printf("No hay ruta para enviar.\n");
	  return -1;
	}

	if(memcmp(best_route->gateway_addr, IPv4_ZERO_ADDR, IPv4_ADDR_SIZE) == 0){ /* Comprobamos si está en la misma red. */
		arp = arp_resolve(ipv4_layer->iface, ipv4_layer->addr, ipv4_addr_dest, mac_dst);
		printf("Enviando a la misma red\n");
	} else {
    arp = arp_resolve(ipv4_layer->iface, ipv4_layer->addr, best_route->gateway_addr, mac_dst);
		printf("Enviando fuera de la red\n");
	}
  if(arp == -1){
    printf("ipv4_send: error de ARP.\n");
		return -1;
	}

	int sent = eth_send(ipv4_layer->iface, mac_dst, IP_ETH_TYPE, (unsigned char *) &packet, total_length);

	if(sent != total_length){
		printf("Error al enviar IP desde Ethernet.\n");
		return -1;
	}
	return (sent - IPv4_HEADER_LENGTH); /* Devuelve la cantidad de datos enviados */
}

/* int ipv4_receive
* 	(	ipv4_layer_t * ipv4, ipv4_addr_t ipv4_addr_src,
*		uint8_t protocol, unsigned char buffer[], int buf_len, long int timeout	);
*
* DESCRIPCIÓN:
*   Esta función recibe la información de la capa superior
*
* PARÁMETROS:
*		'ipv4_layer'		: Estructura ipv4_layer_t con los datos de la máquina que recibe el paquete IP
*   'ipv4_addr_src' : IP origen del envio del paquete IP
*   'protocol'			: Tipo de protocolo.
*   'buffer'				: Array donde se almacenarán los datos de la paquete recibido.
*   'buf_len'				: Longitud del 'buffer' dónde se almacenarán los datos del paquete
*             					recibido. Si se reciben más datos de los que caben el en 'buffer'
*             					sólo se copiaran los primeros 'buf_len' bytes.
*    'timeout' 			: Tiempo en milisegundos que debe esperarse a recibir un paquete
*             					antes de retornar. Un número negativo indicará que debe
*             					esperarse indefinidamente, mientras que con un '0' la función
*             					retornará inmediatamente, se haya recibido o no una trama.
*
* VALOR DEVUELTO:
*   La longitud en bytes de los datos del paquete recibido o '0' si no se ha recibido ninguna
*   trama porque ha expirado el temporizador.
*
* ERRORES:
*   La función devuelve '-1' si se ha producido algún error.
*/

int ipv4_receive( ipv4_layer_t * ipv4_layer, ipv4_addr_t ipv4_addr_src, uint8_t protocol, unsigned char buffer[], int buf_len, long int timeout)
{

	int buf_total_length = IPv4_HEADER_LENGTH + buf_len ; /* Cantidad total de datos a recibir de Ethernet */
	int received = 0;
  ipv4_packet_t * ip_packet_ptr; /* Puntero para acceder a los datos recibidos */
  unsigned char ipv4_buffer[buf_total_length];
	timerms_t timer;
	timerms_reset(&timer, timeout);
	int check_protocol = 0, check_ip = 0, multicast_ip = 0;
	mac_addr_t mac_src;

	do {
	  long int time_left = timerms_left(&timer);
	  received = eth_recv(ipv4_layer->iface, mac_src, IP_ETH_TYPE, ipv4_buffer, buf_total_length, time_left);

	  /* Comprobación de errores */
	  if (received == -1) {
	  	 printf("Error en IPv4 por eth_recv().\n");
	     return -1;
	  } else if (received == 0) {
			printf(" IPv4 Timeout.\n");
			return 0;
		} else if (received < IPv4_HEADER_LENGTH) {
      printf( "IPv4 Paquete de tamaño invalido.\n");
      continue;
    }

		/*Comprobar que el protocolo sea el mismo, que la ip destino es la tuya y el checksum*/
		ip_packet_ptr = (ipv4_packet_t *) ipv4_buffer;
		uint16_t checksum_recv = ntohs(ip_packet_ptr->checksum);
		ip_packet_ptr->checksum = htons(0x0000);
		ip_packet_ptr->checksum = ipv4_checksum((unsigned char *) ip_packet_ptr, IPv4_HEADER_LENGTH);

		if (ip_packet_ptr->checksum != checksum_recv) {
			printf( "IPv4 checksum inválido. id=%d\n", ntohs(ip_packet_ptr->id));
      continue;
		}

  	check_protocol = (ip_packet_ptr->protocol== protocol);
  	check_ip = (memcmp(ip_packet_ptr->dst_addr_ipv4, ipv4_layer->addr, IPv4_ADDR_SIZE) == 0);

		/* La IP es multicast si al hacer addr_dst & 11110000 nos da 224 (224.0.0.0/4) */
	  multicast_ip = (ip_packet_ptr->dst_addr_ipv4[0] & 0xF0) == 0xE0;

	} while ( !(check_protocol && (check_ip || multicast_ip)));

	memcpy(ipv4_addr_src, ip_packet_ptr->src_addr_ipv4, IPv4_ADDR_SIZE);

  int payload_length = ntohs(ip_packet_ptr->total_length) - IPv4_HEADER_LENGTH;

	if (buf_len > payload_length) {
		buf_len = payload_length;
	}

  memcpy(buffer, ip_packet_ptr->payload, buf_len);

	return payload_length;

}

/* void ipv4_getaddr ( ipv4_layer_t * ipv4_layer, ipv4_addr_t addr );
 *
 * DESCRIPCIÓN:
 *   Esta función permite obtener la dirección IP de la máquina especificada.
 *
 * PARÁMETROS:
 *   'ipv4_layer': Estructura ipv4_layer_t con los datos de la máquina
 *            		La estructura debe haber sido inicializada con 'ipv4_open()'
 *            		previamente.
 *    'addr': Array donde se copiará la dirección IP de la estructura ipv4_layer.
 *
 */

void ipv4_getaddr (ipv4_layer_t * ipv4_layer, ipv4_addr_t addr)
{
  if (ipv4_layer != NULL) {
    memcpy(addr, ipv4_layer->addr, IPv4_ADDR_SIZE);
  }
}

/* char * ipv4_getiface_name (ipv4_layer_t * ipv4_layer);
 *
 * DESCRIPCIÓN:
 *   Esta función permite obtener el nombre de la iface de una estructura ipv4_layer_t especificada.
 *
 * PARÁMETROS:
 *   'ipv4_layer': Estructura ipv4_layer_t con los datos de la máquina
 *            		La estructura debe haber sido inicializada con 'ipv4_open()'
 *            		previamente.
 * VALOR DEVUELTO:
 *   Cadena de texto con el nombre de la interfaz.
 *
 */

char * ipv4_getiface_name (ipv4_layer_t * ipv4_layer)
{
	char* iface_name = NULL;

	if (ipv4_layer != NULL) {
		iface_name = eth_getname(ipv4_layer->iface);
  }
	return iface_name;
}
